import requests


def request(verb, url='/', form=None, query=None, files=None):
    kw = {}
    # headers = kw.setdefault('headers', {})

    if form:
        kw['data'] = form

    if query:
        kw['params'] = query

    if files:
        kw['files'] = files

    response = requests.request(verb, url, **kw)

    return response



import os
import re


IGMP_ADDRESS = '224.0.0.70'
IGMP_PORT = 5333
UDP_READSIZE = 128
VERB_DISCOVER = 1
VERB_ANSWER = 2
VERBS = {
    VERB_DISCOVER: 'discover',
    VERB_ANSWER: 'answer'
}
DEFAULT_DBFILE = os.path.join(os.environ['HOME'], '.cache', 'uns')
DEFAULT_TIMEOUT = 5
BINARY_CONTENTTYPES = [
    re.compile('^image/.*'),
]



"""Simple file cache for UNS records."""
from . import protocol


class InvalidDBFileError(Exception):
    pass


class HostNotFoundError(Exception):
    pass


class DB:
    """Simple database for UNS records."""

    def __init__(self, filename):
        self.filename = filename
        self._db = {}
        self._names = {}
        self._dirty = False
        self.load()

    def set(self, addr, hostname, dirty=True):
        self._db[addr] = hostname
        self._names[hostname] = addr
        self._dirty = dirty

    def _parseline(self, l):
        try:
            addr, hostname = l.strip().split(' ')
            self.set(addr, hostname, False)
        except ValueError:
            raise InvalidDBFileError(self.filename)

    def save(self):
        with open(self.filename, 'w') as f:
            for addr, hosts in self._db.items():
                f.write(f'{addr} {hosts}\n')

    def getaddr(self, hostname, resolve=True, resolvetimeout=None):
        addr = self._names.get(hostname)
        if addr:
            return addr, True

        if resolve:
            addr = protocol.resolve(hostname, resolvetimeout)
            self.set(addr, hostname, True)
            return addr, False

        raise HostNotFoundError(hostname)

    def invalidate(self, hostname):
        addr = self._names.get(hostname)
        if not addr:
            return

        del self._names[hostname]
        del self._db[addr]

    def find(self, pattern):
        for addr, host in self._db.items():
            if host.startswith(pattern):
                yield host, addr

    def load(self):
        try:
            with open(self.filename) as f:
                for l in f:
                    self._parseline(l)
        except FileNotFoundError:  # pragma: no cover
            # Don't worry! the file will be created when save() method is
            # called.
            pass

    def __enter__(self):
        return self

    def __exit__(self, ex, extype, tb):
        if self._dirty:
            self.save()



"""UNS.

Simple IGMP discovery protocol
"""
from . import protocol, cache
from .constants import DEFAULT_DBFILE


__version__ = '1.4.0'


def resolve(hostname, timeout=0, force=False, noresolve=False,
            dbfile=DEFAULT_DBFILE):
    assert not (force and noresolve)

    with cache.DB(dbfile) as db:
        if force:
            db.invalidate(hostname)

        return db.getaddr(
            hostname,
            resolve=not noresolve,
            resolvetimeout=timeout
        )



import struct
import socket

from .constants import UDP_READSIZE, IGMP_ADDRESS, IGMP_PORT, VERB_ANSWER, \
    VERB_DISCOVER


def createsocket(timeout=None):
    """Create a udp socket for IGMP multicast."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    if timeout:
        sock.settimeout(timeout)

    return sock


def readpacket(sock):
    data, host = sock.recvfrom(UDP_READSIZE)
    return data[0], data[1:].decode(), host[0], host[1]


def createpacket(verb, data):
    header = struct.pack('>B', verb)
    return header + data.encode()


def resolve(hostname, timeout=None):
    sock = createsocket(timeout)
    sock.sendto(
        createpacket(VERB_DISCOVER, hostname),
        (IGMP_ADDRESS, IGMP_PORT)
    )
    while True:
        verb, name, addr, _ = readpacket(sock)
        if (verb == VERB_ANSWER) and (name == hostname):
            return addr


def sniff():
    sock = createsocket()
    sock.bind((IGMP_ADDRESS, IGMP_PORT))
    mreq = struct.pack(
        '4sl',
        socket.inet_aton(IGMP_ADDRESS),
        socket.INADDR_ANY
    )
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

    while True:
        yield readpacket(sock)


def find(pattern, timeout=None):
    sock = createsocket(timeout)
    sock.sendto(
        createpacket(VERB_DISCOVER, pattern),
        (IGMP_ADDRESS, IGMP_PORT)
    )
    while True:
        verb, name, addr, _ = readpacket(sock)
        if verb == VERB_ANSWER:
            yield name, addr


def answer(name, address=IGMP_ADDRESS, port=IGMP_PORT):
    sock = createsocket()
    sock.sendto(createpacket(VERB_ANSWER, name), (address, port))



import sys
import socket
import functools

import easycli as cli

from .constants import IGMP_PORT, IGMP_ADDRESS, VERBS, DEFAULT_DBFILE, \
    DEFAULT_TIMEOUT, BINARY_CONTENTTYPES
from . import protocol, cache, resolve, http


hostname_arg = functools.partial(cli.Argument, 'hostname')
nocache_arg = functools.partial(cli.Argument, '--nocache', action='store_true')
address_arg = functools.partial(cli.Argument, '-a', '--address')
port_arg = functools.partial(cli.Argument, '-p', '--port', type=int)
short_arg = functools.partial(
    cli.Argument,
    '-s',
    '--short',
    action='store_true'
)
timeout_arg = functools.partial(
    cli.Argument,
    '-t',
    '--timeout',
    type=int,
    default=DEFAULT_TIMEOUT,
    help=f'Seconds wait before exit, 0: infinite, default: {DEFAULT_TIMEOUT}'
)
noresolve_arg = functools.partial(
    cli.Argument,
    '--noresolve',
    action='store_true',
    help='Do not resolve the name over network.'
)
forceresolve_arg = functools.partial(
    cli.Argument,
    '-f', '--forceresolve',
    action='store_true',
    help='Force to resolve the name over network and update cache.'
)


stdout = sys.stdout
stderr = sys.stderr


def output(*a, **kw):
    print(*a, file=stdout, **kw)


def error(*a, **kw):
    print(*a, file=stderr, **kw)


def printrecord(name, addr, cache, short=False):
    if short:
        output(addr)
        return

    flags = ' [cache]' if cache else ''
    output(f'{addr} {name}{flags}')


class Answer(cli.SubCommand):
    """Answer command line interface."""

    __command__ = 'answer'
    __aliases__ = ['a', 'ans']
    __arguments__ = [
        hostname_arg(default=IGMP_ADDRESS),
        address_arg(default=IGMP_ADDRESS, help=f'Default: {IGMP_PORT}'),
        port_arg(default=IGMP_PORT, help=f'Default: {IGMP_ADDRESS}'),
    ]

    def __call__(self, args):
        output(f'Answering {args.hostname} to {args.address}:{args.port}')
        protocol.answer(args.hostname, address=args.address, port=args.port)


class Sniff(cli.SubCommand):
    """Sniff IGMP packets."""

    __command__ = 'sniff'
    __aliases__ = ['s']
    __arguments__ = [
    ]

    def __call__(self, args):
        output(f'Listening to {IGMP_ADDRESS}:{IGMP_PORT}')
        try:
            for verb, name, addr, port in protocol.sniff():
                output(f'{addr}:{port} {VERBS.get(verb)} {name}')
        except KeyboardInterrupt:
            error('Terminated by user.')
            return 3


class Find(cli.SubCommand):
    """Find an ip address by it's name."""

    __command__ = 'find'
    __aliases__ = ['f']
    __arguments__ = [
        cli.Argument('pattern'),
        nocache_arg(),
        short_arg(),
        timeout_arg(),
    ]

    def __call__(self, args):
        if not args.nocache:
            with cache.DB(args.dbfile) as db:
                for name, addr in db.find(args.pattern):
                    printrecord(name, addr, True, short=args.short)

        # Searching network
        for n, a in protocol.find(args.pattern, timeout=args.timeout):
            printrecord(n, a, False, short=args.short)


class Resolve(cli.SubCommand):
    """Resolve an ip address by it's name."""

    __command__ = 'resolve'
    __aliases__ = ['r', 'd']
    __arguments__ = [
        hostname_arg(),
        cli.Mutex(
            noresolve_arg(),
            forceresolve_arg(),
        ),
        short_arg(),
        timeout_arg(),
    ]

    def __call__(self, args):
        addr, cached = resolve(
            args.hostname,
            args.timeout,
            args.forceresolve,
            args.noresolve,
            args.dbfile
        )
        printrecord(args.hostname, addr, cached, short=args.short)


class HTTP(cli.SubCommand):
    """Send HTTP request to a UNS host."""

    __command__ = 'http'
    __aliases__ = ['h']
    __arguments__ = [
        cli.Mutex(
            noresolve_arg(),
            forceresolve_arg(),
        ),
        timeout_arg(),
        cli.Argument(
            '-i', '--include-headers',
            action='store_true',
            help='Include protocol response headers in the output.'
        ),
        port_arg(),
        cli.Argument('verb'),
        cli.Argument('url'),
        cli.Argument(
            'fields',
            metavar='[?]NAME=VALUE',
            default=[],
            nargs='*'
        )
    ]

    def __call__(self, args):
        query = []
        fields = []
        files = []
        body = ''
        path_ = ''
        port = ''

        urlparts = args.url.split('/')
        addr, _ = resolve(
            urlparts[0],
            args.timeout,
            args.forceresolve,
            args.noresolve,
            args.dbfile
        )
        if len(urlparts) > 1:
            path_ = '/' + '/'.join(urlparts[1:])

        if args.port:
            port = f':{args.port}'

        # Fields
        for f in args.fields:
            if '=' not in f:
                body = ' '.join(args.fields)
                break

            k, v = f.split('=')
            if k[0] == '?':
                query.append((k[1:], v))
            elif k[0] == '@':
                files.append((k[1:], open(v)))
            else:
                fields.append((k, v))

        response = http.request(
            args.verb.upper(),
            f'http://{addr}{port}{path_}',
            query=query,
            form=body if body else fields,
            files=files,
        )

        # Handle exception
        if response.status_code >= 400:
            error(response.text, end='')
            return 1

        # Binary output
        if response.headers.get('content-type'):
            for c in BINARY_CONTENTTYPES:
                if c.match(response.headers['content-type']):
                    stdout.buffer.write(response.content)
                    return

        if args.include_headers:
            output(
                '\n'.join([f'{k}: {v}' for k, v in response.headers.items()]),
                end='\n\n'
            )

        output(response.text, end='')


class UNS(cli.Root):
    """UNS root command line handler."""

    __completion__ = True
    __help__ = 'UNS utility.'
    __arguments__ = [
        cli.Argument('-v', '--version', action='store_true'),
        cli.Argument(
            '--dbfile',
            default=DEFAULT_DBFILE,
            help=f'Database file, default: {DEFAULT_DBFILE}'
        ),
        Resolve,
        Answer,
        Sniff,
        Find,
        HTTP,
    ]

    def main(self, *a, **k):
        try:
            return super().main(*a, **k)
        except socket.timeout:
            error('Timeout reached.')
            return 2

        except KeyboardInterrupt:
            error('Terminated by user.')
            return 3

        except cache.InvalidDBFileError as ex:
            error(f'Invalid input file: {ex}')
            return 4

        except cache.HostNotFoundError as ex:
            error(f'Cannot find: {ex}.')
            return 5

    def __call__(self, args):
        if args.version:
            import uns
            output(uns.__version__)
            return

        self._parser.print_help(file=stdout)



"""UNS library command line interface setup script."""
# pylama:ignore=D203,D102
import re

from os.path import join, dirname
from setuptools import setup, find_packages


# reading package version (same way the sqlalchemy does)
with open(join(dirname(__file__), 'uns', '__init__.py')) as v_file:
    package_version = re.compile('.*__version__ = \'(.*?)\'', re.S).\
        match(v_file.read()).group(1)


dependencies = [
    'easycli >= 1.9.1',
    'requests',
]


setup(
    name='uns',
    version=package_version,
    packages=find_packages(),
    install_requires=dependencies,
    include_package_data=True,
    license='MIT',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'uns = uns.cli:UNS.quickstart',
        ]
    }
)
